Option Explicit

' sshd_block.vbs - Blocks IP addresses generating invalid SSH logons
' Copyright 2009 Wellbury LLC - See LICENSE for license information
'
' Release 20090721 - Initial public release

' External executables required to be in PATH:
'
' ROUTE.EXE			For black-hole routing banned IP addresses
' EVENTCREATE.EXE	For writing to the event log (if enabled)
'
' For support, please contact Evan Anderson at Wellbury LLC 
'   EAnderson@wellbury.com, (866) 569-9799, ext 801

' Main
Dim objShell, objWMIService, objEventSink, blackHoleIPAddress, regexpSanitize
Dim dictIPLastSeenTime, dictIPBadLogons, dictUnbanTime, dictBanImmediatelyUsers


' =====================( Configuration )=====================

' Set to 0 to disable debugging output
Const DEBUGGING = 0

' Set to 0 to disable event log reporting of bans / unbans
Const USE_EVENTLOG = 1
Const EVENTLOG_SOURCE = "sshd_block"
Const EVENTLOG_TYPE_INFORMATION = "INFORMATION"
Const EVENTLOG_TYPE_ERROR = "ERROR"
Const EVENTLOG_ID_STARTED = 1
Const EVENTLOG_ID_ERROR_NO_BLACKHOLE_IP = 2
Const EVENTLOG_ID_ERROR_NO_EVENT_MESSAGE_FILE = 3
Const EVENTLOG_ID_BAN = 256
Const EVENTLOG_ID_UNBAN = 257

' Location of Event Message File in registry
Const REGISTRY_EVENT_MESSAGE_FILE = "HKLM\SYSTEM\CurrentControlSet\Services\Eventlog\Application\sshd\EventMessageFile"

' Expiration (in seconds) for IPs to be banned
Const BAN_DURATION = 300

' Number of failed logons in time window before IP will be banner
Const BAN_LIMIT = 5		' Attempts
Const BAN_TIMEOUT = 120	' in X seconds

' Usernames that attempted logons for result in immediate banning
Set dictBanImmediatelyUsers = CreateObject("Scripting.Dictionary")
dictBanImmediatelyUsers.Add "administrator", 1
dictBanImmediatelyUsers.Add "root", 1
dictBanImmediatelyUsers.Add "guest", 1

' ===================( End Configuration )===================


' =====================( Stress Testing )====================

' Set to 1 to perform stress testing
Const TESTING = 0

' Number of "bogus" bans to load
Const TESTING_IP_ADDRESSES = 10000

' Minimum number of milliseconds between adding "bogus" IPs to the ban list during testing
Const TESTING_IP_MIN_LATENCY = 10

' Maximum number of milliseconds between adding "bogus" IPs to the ban list during testing
Const TESTING_IP_MAX_LATENCY = 50

If TESTING Then 
	Dim testLatency, cumulativeLatency, testLoop, maxBanned, bannedAddresses
	Randomize
End IF

' ===================( End Stress Testing )==================

Set dictIPLastSeenTime = CreateObject("Scripting.Dictionary")
Set dictIPBadLogons  = CreateObject("Scripting.Dictionary")
Set dictUnbanTime = CreateObject("Scripting.Dictionary")
Set objShell = CreateObject("WScript.Shell")

Set regexpSanitize = new Regexp
regexpSanitize.Global = True
regexpSanitize.Pattern = "[^0-9a-zA-Z._ /:\-]"

blackHoleIPAddress = GetBlackholeIP()
If IsNull(blackHoleIPAddress) Then
	LogEvent EVENTLOG_ID_ERROR_NO_BLACKHOLE_IP, EVENTLOG_TYPE_ERROR, "Fatal Error - Could not obtain an IP address for an interface with no default gateway specified."
	WScript.Quit
End If

If NOT CheckEventMessageFile() Then
	LogEvent EVENTLOG_ID_ERROR_NO_EVENT_MESSAGE_FILE, EVENTLOG_TYPE_ERROR, "Fatal Error - Event message file for sshd not present."
	WScript.Quit
End IF

' Create event sink to catch sshd events
Set objWMIService = GetObject("winmgmts:{(security)}!root/cimv2")
Set objEventSink = WScript.CreateObject("WbemScripting.SWbemSink", "eventSink_")
objWMIService.ExecNotificationQueryAsync objEventSink, "SELECT * FROM __InstanceCreationEvent WHERE TargetInstance ISA 'Win32_NTLogEvent' AND TargetInstance.SourceName = 'sshd'"

LogEvent EVENTLOG_ID_STARTED, EVENTLOG_TYPE_INFORMATION, EVENTLOG_SOURCE & " started."

If TESTING Then
	If DEBUGGING Then WScript.Echo "Stress test loop"
	
	For testLoop = 1 to TESTING_IP_ADDRESSES 
		testLatency = Int(Rnd() * (TESTING_IP_MAX_LATENCY - TESTING_IP_MIN_LATENCY)) + TESTING_IP_MIN_LATENCY
		
		WScript.Sleep(testLatency)
		Ban(CStr(Int(Rnd * 256)) & "." & CStr(Int(Rnd * 256)) & "." & CStr(Int(Rnd * 256)) & "." & CStr(Int(Rnd * 256)))
		bannedAddresses = bannedAddresses + 1
		
		' Try to ExpireBans no more often than once every 1000ms
		cumulativeLatency = cumulativeLatency + testLatency
		If cumulativeLatency >= 1000 Then
			if bannedAddresses > maxBanned Then maxBanned = bannedAddresses
			cumulativeLatency = 0
			ExpireBans
		End If
	Next ' testLoop

	' Drain the queue
	While dictUnbanTime.Count > 0
		WScript.Sleep(1000)
		ExpireBans
	Wend

	WScript.Echo "Stress test completed. " & TESTING_IP_ADDRESSES & " tested with a maximum of " & maxBanned & " addresses banned at once."

	' Loop until killed
	While (True)
		WScript.Sleep(1000)
	Wend
	
Else
	
	If DEBUGGING Then WScript.Echo "Entering normal operation wait loop."
	
	' Loop sleeping for 1000ms, expiring bans
	While (True)
		WScript.Sleep(1000)
		ExpireBans
	Wend

End If


Sub Ban(IP)
	' Ban an IP address and set the time for the ban expiration
	Dim intRemoveBanTime

	' Ban an IP address by black-hole routing it
	If (TESTING <> 1) Then objShell.Run "route add " & IP & " mask 255.255.255.255 " & blackHoleIPAddress 

	' Calculate time to remove ban and add to dictUnbanTime
	intRemoveBanTime = (Date + Time) + (BAN_DURATION / (24 * 60 * 60))

	If NOT dictUnbanTime.Exists(intRemoveBanTime) Then
		Set dictUnbanTime.Item(intRemoveBanTime) = CreateObject("Scripting.Dictionary")
	End If
	If NOT dictUnbanTime.Item(intRemoveBanTime).Exists(IP) Then dictUnbanTime.Item(intRemoveBanTime).Add IP, 1

	LogEvent EVENTLOG_ID_BAN, EVENTLOG_TYPE_INFORMATION, "Banned " & IP & " until " & intRemoveBanTime
End Sub

Sub Unban(IP)
	' Unban an IP address
	If (TESTING <> 1) Then objShell.Run "route delete " & IP & " mask 255.255.255.255 " & blackHoleIPAddress 
	
	LogEvent EVENTLOG_ID_UNBAN, EVENTLOG_TYPE_INFORMATION, "Unbanned " & IP
End Sub

Sub LogFailedLogonAttempt(IP)
	' Log failed logon attempts and, if necessary, ban the IP address
	
	' Have we already seen this IP address before?
	If dictIPLastSeenTime.Exists(IP) Then

		' Be sure that prior attempts, if they are older than BAN_TIMEOUT, don't count it against the IP
		If (dictIPLastSeenTime.Item(IP) + (BAN_TIMEOUT / (24 * 60 * 60))) <= (Date + Time) Then
			If dictIPBadLogons.Exists(IP) Then dictIPBadLogons.Remove(IP)
		End If
		
		dictIPLastSeenTime.Item(IP) = (Date + Time)
	Else
		dictIPLastSeenTime.Add IP, (Date + Time)
	End If
	
	' Does this IP address already have a history of bad logons?
	If dictIPBadLogons.Exists(IP) Then
		dictIPBadLogons.Item(IP) = dictIPBadLogons.Item(IP) + 1
	Else
		dictIPBadLogons.Add IP, 1
	End If

	If DEBUGGING Then WScript.Echo "Logging bad attempt from " & IP & ", attempt # " & dictIPBadLogons.Item(IP)

	' Should we ban this IP address?
	If dictIPBadLogons.Item(IP) = BAN_LIMIT Then Ban(IP)
End Sub

Sub ExpireBans()
	Dim unbanTime, ipAddress
	
	For Each unbanTime in dictUnbanTime.Keys

		If unbanTime <= (Date + Time) Then 
			For Each ipAddress in dictUnbanTime.Item(unbanTime)
				Unban(ipAddress)
				If TESTING Then bannedAddresses = bannedAddresses - 1
			Next ' ipAddress
			
			dictUnbanTime.Remove unbanTime
		End If
	Next 'ipAddress
End Sub

' Should an invalid logon from specified user result in an immediate ban?
Function BanImmediate(user)
	Dim userToBan
	
	For Each userToBan in dictBanImmediatelyUsers.Keys
		If UCase(user) = UCase(userToBan) Then 
			BanImmediate = True
			Exit Function
		End If
	Next 'userToBan
	
	BanImmediate = False
End Function

' Fires each time new OpenSSHd events are generated
Sub eventSink_OnObjectReady(objEvent, objWbemAsyncContext)
	Dim arrEventMessage, arrInvalidLogonText
	Dim IP, user
	
	arrEventMessage = Split(objEvent.TargetInstance.Message, ":")
	
	If Left(LTrim(UCase(arrEventMessage(2))), 12) = "ILLEGAL USER" Then
		arrInvalidLogonText = Split(LTrim(arrEventMessage(2)), " ")
		user = arrInvalidLogonText(2)
		IP = arrInvalidLogonText(4)
	ElseIf Left(LTrim(UCase(arrEventMessage(2))), 32) = "FAILED PASSWORD FOR ILLEGAL USER" Then
		arrInvalidLogonText = Split(LTrim(arrEventMessage(2)), " ")
		user = arrInvalidLogonText(5)
		IP = arrInvalidLogonText(7)
	ElseIf Left(LTrim(UCase(arrEventMessage(2))), 19) = "FAILED PASSWORD FOR" Then
		arrInvalidLogonText = Split(LTrim(arrEventMessage(2)), " ")
		user = arrInvalidLogonText(3)
		IP = arrInvalidLogonText(5)
	End If
	
	' If the event didn't generate both a username and IP address then do nothing
	If (IP <> "") AND (user <> "") Then
		If BanImmediate(user) Then Ban(IP) Else LogFailedLogonAttempt(IP)
	End If
End Sub

Function GetBlackholeIP()
	' Sift through the NICs on the machine to locate a NIC's IP to use to blackhole offending hosts.
	' Look for a NIC with no default gateway set and an IP address assigned. Return NULL if we can't
	' find one.
	
	Dim objNICs, objNICConfig
	Set objNICs = GetObject("winmgmts:\\.\root\cimv2").ExecQuery("SELECT * FROM Win32_NetworkAdapterConfiguration WHERE IPEnabled = TRUE")
	
	' Scan for a NIC with no default gateway set and IP not 0.0.0.0
	For Each objNICConfig in objNICs
		If IsNull(objNICConfig.DefaultIPGateway) and (objNICConfig.IPAddress(0) <> "0.0.0.0") Then 
			If DEBUGGING Then WScript.Echo "Decided on black-hole IP address " & objNICConfig.IPAddress(0) & ", interface " & objNICConfig.Description
			GetBlackholeIP = objNICConfig.IPAddress(0)
			Exit Function
		End If
	Next

	' Couldn't find anything, return NULL to let caller know we failed
	GetBlackHoleIP = NULL
End Function

Sub LogEvent(ID, EventType, Message)
	' Log an event to the Windows event log
	
	' Sanitize input string
	Message = regexpSanitize.Replace(Message, "")
	
	If DEBUGGING Then WScript.Echo "Event Log - Event ID: " & ID & ", Type: " & EventType & " - " & Message
	If USE_EVENTLOG Then objShell.Exec "EVENTCREATE /L APPLICATION /SO " & EVENTLOG_SOURCE & " /ID " & ID & " /T " & EventType & " /D """ & Message & """"
End Sub

Function CheckEventMessageFile()
	Dim eventMessageFile
	On Error Resume Next ' Noooooooo!
	
	eventMessageFile = objShell.RegRead(REGISTRY_EVENT_MESSAGE_FILE)
	
	If Err.Number <> 0 Then CheckEventMessageFile = False Else CheckEventMessageFile = True
	
	On Error Goto 0
End Function